/*******************************************************************************
This software module was originally developed by

Agere Systems, Coding Technologies, Fraunhofer IIS, Philips

and edited by

-

in the course of development of ISO/IEC 23003 for reference purposes and its
performance may not have been optimized. This software module is an
implementation of one or more tools as specified by ISO/IEC 23003. ISO/IEC gives
You a royalty-free, worldwide, non-exclusive, copyright license to copy,
distribute, and make derivative works of this software module or modifications
thereof for use in implementations of ISO/IEC 23003 in products that satisfy
conformance criteria (if any). Those intending to use this software module in
products are advised that its use may infringe existing patents. ISO/IEC have no
liability for use of this software module or modifications thereof. Copyright is
not released for products that do not conform to audiovisual and image-coding
related ITU Recommendations and/or ISO/IEC International Standards.

Agere Systems, Coding Technologies, Fraunhofer IIS, Philips retain full right to
modify and use the code for its (their) own purpose, assign or donate the code
to a third party and to inhibit third parties from using the code for products
that do not conform to MPEG-related ITU Recommendations and/or ISO/IEC
International Standards. This copyright notice must be included in all copies or
derivative works.

Copyright (c) ISO/IEC 2008.
*******************************************************************************/
/*******************************************************************************
This software module was further modified in light of the JAME project that 
was initiated by Yonsei University and LG Electronics.

The JAME project aims at making a better MPEG Reference Software especially
for the encoder in direction to simplification, quality improvement and 
efficient implementation being still compliant with the ISO/IEC specification. 
Those intending to participate in this project can use this software module 
under consideration of above mentioned ISO/IEC copyright notice. 

Any voluntary contributions to this software module in the course of this 
project can be recognized by proudly listing up their authors here:

- Henney Oh, LG Electronics (henney.oh@lge.com)
- Sungyong Yoon, LG Electronics (sungyong.yoon@lge.com)

Copyright (c) JAME 2010.
*******************************************************************************/
                                                                 
#include "sac_huff_tab.h"                                       


const HUFF_CLD_TABLE huffCLDTab =
{
  {
    { 0x52, 0xae, 0xaf, 0x28, 0x6e, 0x36, 0x1e, 0x0e, 0x0c, 0x0a, 0x02, 0x16, 0x12, 0x17, 0x00, 0x04, 0x06, 0x08, 0x07, 0x03, 0x01, 0x1a, 0x13, 0x3e, 0x16, 0x17, 0x6f, 0x2a, 0x56, 0x53, 0x3f },
    {    8,    9,    9,    7,    7,    6,    5,    4,    4,    4,    4,    5,    5,    5,    4,    4,    4,    4,    4,    4,    4,    5,    5,    6,    6,    6,    7,    7,    8,    8,    6 }
  },
  {
    {
      { 0x0000, 0x0002, 0x0006, 0x000e, 0x001e, 0x003e, 0x007e, 0x00fe, 0x01fe, 0x03fe, 0x07fe, 0x0ffe, 0x1ffe, 0x7ffe, 0x7ffc, 0xfffe, 0xfffa, 0x1fffe, 0x1fff6, 0x3fffe, 0x3ffff, 0x7ffde, 0x3ffee, 0xfffbe, 0x1fff7e, 0xfffbfc, 0xfffbfd, 0xfffbfe, 0xfffbff, 0x7ffdfc, 0x7ffdfd },
      {      1,      2,      3,      4,      5,      6,      7,      8,      9,     10,     11,     12,     13,     15,     15,     16,     16,     17,     17,     18,     18,     19,     18,     20,     21,     24,     24,     24,     24,     23,     23 }
    },
    {
      { 0x0000, 0x0002, 0x0006, 0x000e, 0x001e, 0x003e, 0x007e, 0x01fe, 0x01fc, 0x03fe, 0x03fa, 0x07fe, 0x07f6, 0x0ffe, 0x0fee, 0x1ffe, 0x1fde, 0x3ffe, 0x3fbe, 0x3fbf, 0x7ffe, 0xfffe, 0x1fffe, 0x7fffe, 0x7fffc, 0xffffa, 0x1ffffc, 0x1ffffd, 0x1ffffe, 0x1fffff, 0xffffb },
      {      1,      2,      3,      4,      5,      6,      7,      9,      9,     10,     10,     11,     11,     12,     12,     13,     13,     14,     14,     14,     15,     16,     17,     19,     19,     20,     21,     21,     21,     21,     20 }
    }
  },
  { /* escape */
    { 
      { 0x0, 0x0, 0x0, 0x0 },
      {   0,   0,   0,   0 }
    },
    {
      { 0x0, 0x0, 0x17feff, 0x17fef7 },
      {   0,   0,       21,       21 }
    },
    {
      { 0x0, 0x0, 0x0, 0x0b7dff },
      {   0,   0,   0,       21 }
    },
    {
      { 0x0, 0x0, 0x0, 0x0 },
      {   0,   0,   0,   0 }
    }
  },
  {
    { /* lav3 */
      {
        { 0x0002, 0x0002, 0x0004, 0x003e, 0x0006, 0x0007, 0x000e, 0x00fe, 0x007e, 0x001e, 0x000c, 0x0005, 0x00ff, 0x000d, 0x0000, 0x0003, },
        {      2,      3,      5,      8,      4,      4,      6,     10,      9,      7,      6,      5,     10,      6,      3,      2, }
      },
      {
        { 0x0000, 0x003e, 0x076e, 0x0ede, 0x0006, 0x003f, 0x03b6, 0x003a, 0x001c, 0x00ee, 0x01da, 0x001e, 0x00ef, 0x0edf, 0x00ec, 0x0002, },
        {      1,      6,     11,     12,      3,      6,     10,      6,      5,      8,      9,      5,      8,     12,      8,      2, }
      },
      {
        { 0x0000, 0x001e, 0x03be, 0x0efe, 0x0006, 0x001c, 0x01de, 0x00ea, 0x0074, 0x00ee, 0x00eb, 0x001f, 0x077e, 0x0eff, 0x0076, 0x0002, },
        {      1,      5,     10,     12,      3,      5,      9,      8,      7,      8,      8,      5,     11,     12,      7,      2, }
      },
      {
        { 0x0000, 0x000e, 0x00fa, 0x07de, 0x000c, 0x001e, 0x00fe, 0x01f6, 0x00ff, 0x007c, 0x007e, 0x001a, 0x07df, 0x03ee, 0x001b, 0x0002, },
        {      1,      4,      8,     11,      4,      5,      8,      9,      8,      7,      7,      5,     11,     10,      5,      2, }
      }
    },
    { /* lav5 */
      {
        { 0x0002, 0x0003, 0x0010, 0x007c, 0x00d6, 0x03ee, 0x000a, 0x000c, 0x0016, 0x0034, 0x00fe, 0x1f7e, 0x007e, 0x0036, 0x0026, 0x0046, 0x011e, 0x01f6, 0x011f, 0x00d7, 0x008e, 0x00ff, 0x006a, 0x004e, 0x0fbe, 0x07de, 0x004f, 0x0037, 0x0017, 0x001e, 0x1f7f, 0x00fa, 0x0022, 0x0012, 0x000e, 0x0000, },
        {      3,      3,      5,      7,      8,     10,      4,      4,      5,      6,      8,     13,      7,      6,      6,      7,      9,      9,      9,      8,      8,      8,      7,      7,     12,     11,      7,      6,      5,      5,     13,      8,      6,      5,      4,      2, }
      },
      {
        { 0x0006, 0x001c, 0x007e, 0x0efc, 0xeffe, 0x1dffe, 0x0004, 0x000a, 0x003e, 0x0efe, 0x77fe, 0x0076, 0x0006, 0x0016, 0x00be, 0x0efd, 0x00ee, 0x000e, 0x003e, 0x002e, 0x01de, 0x03be, 0x007e, 0x001e, 0x007f, 0x005e, 0x3bfe, 0x00fe, 0x001e, 0x0002, 0x00bf, 0x1dfff, 0x1dfe, 0x00ff, 0x003a, 0x0000, },
        {      3,      5,      8,     12,     16,     17,      3,      4,      7,     12,     15,      7,      4,      5,      8,     12,      8,      5,      6,      6,      9,     10,      7,      5,      7,      7,     14,      9,      6,      3,      8,     17,     13,      9,      6,      2, }
      },
      {
        { 0x0000, 0x0006, 0x0024, 0x025e, 0x3cfe, 0x79fe, 0x0006, 0x0007, 0x0078, 0x03ce, 0x1e7e, 0x00be, 0x0008, 0x003e, 0x0026, 0x012e, 0x00bf, 0x002e, 0x0027, 0x007a, 0x01e4, 0x0096, 0x007b, 0x003f, 0x01e6, 0x01e5, 0x0f3e, 0x005e, 0x0016, 0x000e, 0x079e, 0x79ff, 0x025f, 0x004a, 0x000a, 0x0002, },
        {      2,      4,      7,     11,     14,     15,      3,      4,      7,     10,     13,      9,      5,      6,      7,     10,      9,      7,      7,      7,      9,      9,      7,      6,      9,      9,     12,      8,      6,      4,     11,     15,     11,      8,      5,      2, }
      },
      {
        { 0x0006, 0x000e, 0x007c, 0x03fe, 0x0fbe, 0x3efe, 0x0000, 0x0001, 0x003c, 0x005e, 0x07de, 0x07be, 0x001e, 0x000a, 0x001f, 0x005f, 0x01ee, 0x01f6, 0x01fe, 0x00fe, 0x00f6, 0x00fa, 0x007e, 0x0016, 0x07bf, 0x03de, 0x03ee, 0x007a, 0x000e, 0x0006, 0x3eff, 0x1f7e, 0x03ff, 0x002e, 0x0004, 0x0002, },
        {      3,      4,      7,     10,     12,     14,      3,      3,      6,      8,     11,     11,      6,      5,      6,      8,      9,      9,      9,      8,      8,      8,      7,      6,     11,     10,     10,      7,      5,      4,     14,     13,     10,      7,      4,      2, }
      }
    },
    { /* lav7 */
      {
        { 0x000e, 0x000a, 0x000a, 0x007c, 0x00be, 0x017a, 0x00ee, 0x07b6, 0x0006, 0x000c, 0x0016, 0x0026, 0x003e, 0x002e, 0x01ec, 0x47ce, 0x0016, 0x003c, 0x0022, 0x004e, 0x003f, 0x005e, 0x08fa, 0x08fb, 0x005f, 0x00fa, 0x00bf, 0x003a, 0x01f6, 0x01de, 0x03da, 0x07b7, 0x01df, 0x03ee, 0x017b, 0x03ef, 0x01ee, 0x008e, 0x01ef, 0x01fe, 0x08f8, 0x047e, 0x047f, 0x0076, 0x003c, 0x0046, 0x007a, 0x007e, 0x23e6, 0x11f2, 0x01ff, 0x003d, 0x004f, 0x002e, 0x0012, 0x0004, 0x47cf, 0x011e, 0x00bc, 0x00fe, 0x001c, 0x0010, 0x000d, 0x0000, },
        {      4,      4,      5,      7,      8,      9,      9,     11,      4,      4,      5,      6,      7,      7,      9,     15,      6,      6,      6,      7,      7,      8,     12,     12,      8,      8,      8,      7,      9,     10,     10,     11,     10,     10,      9,     10,      9,      8,      9,      9,     12,     11,     11,      8,      7,      7,      7,      7,     14,     13,      9,      7,      7,      6,      5,      4,     15,      9,      8,      8,      6,      5,      4,      2, }
      },
      {
        { 0x0002, 0x001c, 0x00bc, 0x05fc, 0x5ffe, 0x2ffde, 0xbff7e, 0x17feff, 0x0004, 0x000a, 0x000e, 0x02fa, 0x01fe, 0xbff2, 0x5ffbe, 0x00ee, 0x0002, 0x0016, 0x00f6, 0x05fe, 0x01ff, 0xbff6, 0x01de, 0x007e, 0x0000, 0x003c, 0x000e, 0x003e, 0x2ffe, 0x02fb, 0x00f7, 0x002e, 0x0006, 0x007a, 0x000a, 0x007e, 0x00fe, 0x0016, 0x0006, 0x0002, 0x000f, 0x0076, 0x0017, 0x5ff8, 0x0bfe, 0x001e, 0x007f, 0x0003, 0x0004, 0x00bd, 0xbff3, 0x5fff, 0x0bfa, 0x017c, 0x003a, 0x0003, 0x017e, 0x17fefe, 0x17fee, 0x5ffa, 0x0bfb, 0x01df, 0x003e, 0x0006, },
        {      3,      5,      8,     11,     15,     18,     20,     21,      3,      4,      7,     10,     13,     16,     19,      8,      4,      5,      8,     11,     13,     16,      9,      7,      5,      6,      8,     10,     14,     10,      8,      6,      6,      7,      8,     11,     12,      9,      7,      5,      7,      7,      9,     15,     12,      9,      7,      4,      7,      8,     16,     15,     12,      9,      6,      3,      9,     21,     17,     15,     12,      9,      6,      3, }
      },
      {
        { 0x0000, 0x0006, 0x00de, 0x069e, 0x34fe, 0x1a7fe, 0x69ff6, 0x69ff7, 0x0002, 0x000c, 0x006a, 0x034e, 0x1fde, 0x69fe, 0x1a7fc, 0x0372, 0x003e, 0x003c, 0x00df, 0x01ee, 0x0dde, 0x69fa, 0x0373, 0x007a, 0x003e, 0x0068, 0x01ba, 0x03f6, 0x0d3e, 0x034c, 0x01fa, 0x00d2, 0x007e, 0x007f, 0x01f8, 0x06ee, 0x03de, 0x01b8, 0x01fc, 0x006b, 0x00f6, 0x01fe, 0x034d, 0x3fbe, 0x07f6, 0x03fa, 0x003c, 0x003d, 0x03f7, 0x0376, 0x1a7ff, 0x3fbf, 0x0ddf, 0x01f9, 0x0036, 0x000e, 0x03df, 0x34ffa, 0x69fb, 0x34fc, 0x0fee, 0x01ff, 0x000e, 0x0002, },
        {      2,      4,      8,     11,     14,     17,     19,     19,      3,      4,      7,     10,     13,     15,     17,     10,      6,      6,      8,     10,     12,     15,     10,      8,      7,      7,      9,     10,     12,     10,      9,      8,      8,      8,      9,     11,     11,      9,      9,      7,      9,      9,     10,     14,     11,     10,      7,      6,     10,     10,     17,     14,     12,      9,      6,      4,     11,     18,     15,     14,     12,      9,      5,      2, }
      },
      {
        { 0x0002, 0x000a, 0x001a, 0x01be, 0x06e6, 0x067a, 0x0cf2, 0x33de, 0x000c, 0x000e, 0x000e, 0x00de, 0x0372, 0x03d6, 0x0678, 0x0cf6, 0x0036, 0x0012, 0x003e, 0x003c, 0x01b8, 0x03d4, 0x033e, 0x033f, 0x007e, 0x006a, 0x004e, 0x007e, 0x01ba, 0x00ce, 0x00f6, 0x01ee, 0x01ef, 0x013e, 0x007f, 0x0066, 0x00d6, 0x003e, 0x00d7, 0x009e, 0x07ae, 0x01e8, 0x01e9, 0x027e, 0x0032, 0x0018, 0x0026, 0x0034, 0x0cf3, 0x07aa, 0x07ab, 0x027f, 0x01bf, 0x001b, 0x001e, 0x000b, 0x33df, 0x19ee, 0x07af, 0x06e7, 0x01bb, 0x007f, 0x0008, 0x0000, },
        {      3,      4,      6,      9,     11,     12,     13,     15,      4,      4,      5,      8,     10,     11,     12,     13,      6,      5,      6,      7,      9,     11,     11,     11,      8,      7,      7,      7,      9,      9,      9,     10,     10,      9,      8,      8,      8,      7,      8,      8,     12,     10,     10,     10,      7,      6,      6,      6,     13,     12,     12,     10,      9,      6,      5,      4,     15,     14,     12,     11,      9,      7,      4,      2, }
      }
    },
    { /* lav 9 */
      {
        { 0x0006, 0x0007, 0x0006, 0x007e, 0x000a, 0x001e, 0x008a, 0x004e, 0x0276, 0x02e2, 0x0000, 0x000a, 0x0016, 0x0026, 0x0076, 0x00f2, 0x0012, 0x005e, 0x008b, 0x2e76, 0x0012, 0x0007, 0x0038, 0x007c, 0x0008, 0x0046, 0x00f6, 0x01ca, 0x173a, 0x1738, 0x009e, 0x004a, 0x0026, 0x000c, 0x004e, 0x00f7, 0x013a, 0x009e, 0x09fe, 0x013e, 0x0026, 0x001a, 0x01e6, 0x01e2, 0x00ee, 0x01ce, 0x0277, 0x03ce, 0x02e6, 0x04fc, 0x02e3, 0x0170, 0x0172, 0x00ba, 0x003e, 0x01e3, 0x001b, 0x003f, 0x009e, 0x009f, 0x0b9e, 0x09ff, 0x04fd, 0x04fe, 0x01cf, 0x00ef, 0x0044, 0x005f, 0x00e4, 0x00f0, 0x2e72, 0x013f, 0x0b9f, 0x013e, 0x00fe, 0x0047, 0x000e, 0x007d, 0x0010, 0x0024, 0x2e77, 0x5ce6, 0x00bb, 0x00e6, 0x0016, 0x00ff, 0x007a, 0x003a, 0x0017, 0x0002, 0x5ce7, 0x03cf, 0x0017, 0x01cb, 0x009c, 0x004b, 0x0016, 0x000a, 0x0008, 0x0006, },
        {      4,      4,      5,      7,      7,      8,      9,     10,     10,     11,      4,      4,      5,      6,      7,      8,      8,      8,      9,     15,      6,      5,      6,      7,      7,      8,      8,      9,     14,     14,      8,      7,      7,      7,      8,      8,      9,     11,     12,     12,      9,      8,      9,      9,      8,      9,     10,     10,     11,     11,     11,     10,     10,      9,      9,      9,      8,      9,      9,      9,     13,     12,     11,     11,      9,      8,      8,      8,      8,      8,     15,     12,     13,      9,      8,      8,      7,      7,      6,      6,     15,     16,      9,      8,      8,      8,      7,      6,      5,      4,     16,     10,      8,      9,      8,      7,      6,      5,      4,      3, }
      },
      {
        { 0x000e, 0x0014, 0x008e, 0x04fe, 0x23fe, 0x8ffe, 0x5ffbc, 0x17fef7, 0x17fef7, 0x17fef7, 0x0002, 0x0002, 0x0044, 0x027e, 0x17fc, 0xbff6, 0x5ffbe, 0x11ff8, 0xbff7a, 0x00bc, 0x0006, 0x0016, 0x001a, 0x00fe, 0x11f6, 0xbffe, 0x11ff9, 0x17fef6, 0x011e, 0x0056, 0x0010, 0x003e, 0x009e, 0x07fe, 0x11f7, 0x5ff8, 0x17fee, 0x07ff, 0x00ae, 0x001e, 0x0026, 0x000e, 0x01ee, 0x047e, 0x0bfc, 0xbfff, 0x08fa, 0x006e, 0x01ef, 0x007e, 0x007a, 0x004e, 0x007e, 0x00de, 0x11fe, 0x2ffe, 0x04ff, 0x00ff, 0x00bd, 0x002e, 0x00fe, 0x00af, 0x01ec, 0x01be, 0x11ffe, 0x2ffa, 0x08fe, 0x03fe, 0x0046, 0x0012, 0x003e, 0x0045, 0x02fe, 0xbff7e, 0x5ff9, 0x5ffa, 0x0bfd, 0x013e, 0x000c, 0x0007, 0x00be, 0x0036, 0xbff7f, 0x23ffe, 0x11ffa, 0x5ffe, 0x01bf, 0x01ed, 0x002a, 0x0000, 0x017e, 0x17fef7, 0x47ffe, 0x47fff, 0x11ffb, 0x2ffb, 0x047c, 0x01fe, 0x003c, 0x0006, },
        {      4,      5,      8,     11,     14,     16,     19,     21,     21,     21,      3,      4,      7,     10,     13,     16,     19,     17,     20,      8,      4,      5,      7,     10,     13,     16,     17,     21,      9,      7,      5,      6,      8,     11,     13,     15,     17,     11,      8,      7,      6,      6,      9,     11,     12,     16,     12,      9,      9,      7,      7,      7,      9,     10,     13,     14,     11,     10,      8,      6,      8,      8,      9,     11,     17,     14,     12,     10,      7,      5,      8,      7,     10,     20,     15,     15,     12,      9,      6,      4,      8,      8,     20,     18,     17,     15,     11,      9,      6,      3,      9,     21,     19,     19,     17,     14,     11,      9,      6,      3, }
      },
      {
        { 0x0006, 0x0004, 0x0012, 0x07fe, 0x1f7e, 0xfbfe, 0x1f7fe, 0xb7dfe, 0xb7dff, 0xb7dff, 0x0000, 0x0006, 0x007c, 0x0046, 0x07d0, 0x1f4e, 0xb7fe, 0x5bee, 0x16fbe, 0x03ee, 0x0006, 0x000a, 0x002e, 0x03fe, 0x07d2, 0x1f4f, 0x2dfe, 0xb7de, 0x01fe, 0x002e, 0x007a, 0x007e, 0x007a, 0x01fa, 0x07fe, 0x1f7c, 0x16fa, 0x009e, 0x0020, 0x0021, 0x00fe, 0x0016, 0x00fe, 0x016e, 0x009f, 0x0b7c, 0x03de, 0x00b6, 0x00be, 0x007c, 0x005a, 0x0078, 0x0047, 0x0044, 0x07ff, 0x07d1, 0x01f6, 0x01f7, 0x002f, 0x002c, 0x00fc, 0x01f6, 0x00f6, 0x07ff, 0x16fe, 0x02de, 0x03ea, 0x00bf, 0x00fa, 0x000a, 0x004e, 0x0026, 0x01ee, 0x5bfe, 0x3efe, 0x0b7e, 0x03eb, 0x01fe, 0x007b, 0x0007, 0x01fb, 0x0045, 0x16ffe, 0x1f7ff, 0x2df6, 0x1f7d, 0x03fe, 0x005e, 0x003c, 0x000e, 0x03df, 0x5befe, 0x2df7e, 0x16fff, 0x7dfe, 0x0fa6, 0x07de, 0x0079, 0x000e, 0x0002, },
        {      3,      4,      7,     11,     13,     16,     17,     21,     21,     21,      3,      4,      7,      9,     12,     14,     17,     16,     18,     10,      5,      5,      7,     10,     12,     14,     15,     17,     10,      8,      7,      7,      8,     10,     12,     13,     14,     10,      8,      8,      8,      7,      9,     10,     10,     13,     11,      9,      9,      8,      8,      8,      9,      9,     12,     12,     10,     10,      8,      7,      9,      9,      9,     11,     14,     11,     11,      9,      8,      6,      9,      8,     10,     16,     14,     13,     11,      9,      7,      5,     10,      9,     18,     17,     15,     13,     11,      8,      6,      4,     11,     20,     19,     18,     15,     13,     11,      8,      5,      2, }
      },
      {
        { 0x000e, 0x0008, 0x007e, 0x01fe, 0x01ba, 0x0dbe, 0x0d7e, 0x1af6, 0x7fec, 0x1ffb6, 0x000a, 0x000c, 0x000c, 0x0036, 0x00de, 0x05fe, 0x06be, 0x1b7e, 0x7fee, 0x6dfe, 0x001e, 0x000e, 0x000a, 0x006a, 0x01ae, 0x06fe, 0x0376, 0x0dfe, 0x0dff, 0x0d7f, 0x00b6, 0x005e, 0x007c, 0x006e, 0x006a, 0x016a, 0x0ffe, 0x0dfe, 0x0ffc, 0x1bfe, 0x035e, 0x01b6, 0x005e, 0x00b4, 0x006c, 0x017e, 0x036e, 0x03ee, 0x037e, 0x0377, 0x0fff, 0x01ae, 0x01be, 0x01f6, 0x01be, 0x00da, 0x00fe, 0x016b, 0x00d6, 0x037e, 0x17fe, 0x0bfe, 0x07de, 0x06de, 0x01b8, 0x00d6, 0x002e, 0x0034, 0x00de, 0x00be, 0x7fef, 0x06bc, 0x1bff, 0x1ffa, 0x01b9, 0x03fe, 0x00fa, 0x002e, 0x0034, 0x001f, 0x6dff, 0x1af7, 0x36fe, 0x06fe, 0x0fbe, 0x035f, 0x00b7, 0x002c, 0x001e, 0x0009, 0x1ffb7, 0xffda, 0x0d7a, 0x17ff, 0x0fbf, 0x02fe, 0x005f, 0x0016, 0x0004, 0x0000, },
        {      4,      4,      7,      9,     10,     12,     13,     14,     15,     17,      4,      4,      5,      7,      9,     11,     12,     13,     15,     15,      6,      5,      5,      7,      9,     11,     11,     13,     13,     13,      8,      7,      7,      7,      8,      9,     12,     12,     12,     13,     10,      9,      8,      8,      7,      9,     10,     10,     11,     11,     12,     10,     10,      9,      9,      8,      8,      9,      9,     10,     13,     12,     11,     11,     10,      8,      7,      7,      8,      8,     15,     12,     13,     13,     10,     10,      8,      6,      6,      6,     15,     14,     14,     12,     12,     10,      8,      6,      5,      4,     17,     16,     13,     13,     12,     10,      8,      6,      4,      2, }
      }
    }
  }
};
